from __future__ import annotations

from abc import ABC


class TodomanError(Exception, ABC):
    """
    Base class for all our exceptions.

    Should not be raised directly.
    """

    EXIT_CODE: int


class NoSuchTodoError(TodomanError):
    EXIT_CODE = 20

    def __str__(self) -> str:
        return f"No todo with id {self.args[0]}."


class ReadOnlyTodoError(TodomanError):
    EXIT_CODE = 21

    def __str__(self) -> str:
        return (
            "Todo is in read-only mode because there are multiple todos "
            f"in {self.args[0]}."
        )


class NoListsFoundError(TodomanError):
    EXIT_CODE = 22

    def __str__(self) -> str:
        return (
            f"No lists found matching {self.args[0]}, create a "
            "directory for a new list."
        )


class AlreadyExistsError(TodomanError):
    """
    Raised when two objects have a same identity.

    This can ocurrs when two lists have the same name, or when two Todos have
    the same path.
    """

    EXIT_CODE = 23

    def __str__(self) -> str:
        return "More than one {} has the same identity: {}.".format(*self.args)



from __future__ import annotations

import logging
import os
import socket
import sqlite3
from datetime import date
from datetime import datetime
from datetime import time
from datetime import timedelta
from datetime import timezone
from functools import cached_property
from os.path import normpath
from os.path import split
from typing import Iterable
from typing import Iterator
from uuid import uuid4

import icalendar
from atomicwrites import AtomicWriter
from dateutil.rrule import rrulestr
from dateutil.tz import tzlocal

from todoman import exceptions

logger = logging.getLogger(name=__name__)

# Initialize this only once
# We were doing this all over the place (even if unused!), so at least only do
# it once.
LOCAL_TIMEZONE = tzlocal()


class Todo:
    """
    Represents a task/todo, and wrapps around icalendar.Todo.

    All text attributes are always treated as text, and "" will be returned if
    they are not defined.
    Date attributes are treated as datetime objects, and None will be returned
    if they are not defined.
    All datetime objects have tzinfo, either the one defined in the file, or
    the local system's one.
    """

    categories: list[str]
    completed_at: datetime | None
    created_at: datetime | None
    due: date | None
    dtstamp: datetime | None
    last_modified: datetime | None
    related: list[Todo]
    rrule: str | None
    start: date | None
    id: int | None

    def __init__(
        self,
        filename: str | None = None,
        mtime: int | None = None,
        new: bool = False,
        list: TodoList | None = None,
    ) -> None:
        """Creates a new todo using `todo` as a source.

        :param filename: The name of the file for this todo. Defaults to the <uid>.ics
        :param int: The last modified time for the file backing this Todo.
        :param new: Indicate that a new Todo is being created and should be populated
            with default values.
        :param list: The list to which this Todo belongs.
        """
        self.list = list
        now = datetime.now(LOCAL_TIMEZONE)
        self.uid = f"{uuid4().hex}@{socket.gethostname()}"

        if new:
            self.created_at = now
        else:
            self.created_at = None

        # Default values for supported fields
        self.categories = []
        self.completed_at = None
        self.description = ""
        self.dtstamp = now
        self.due = None
        self.id = None
        self.last_modified = None
        self.location = ""
        self.percent_complete = 0
        self.priority = 0
        self.rrule = ""
        self.sequence = 0
        self.start = None
        self.status = "NEEDS-ACTION"
        self.summary = ""

        self.filename = filename or f"{self.uid}.ics"
        self.related = []

        if os.path.basename(self.filename) != self.filename:
            raise ValueError(f"Must not be an absolute path: {self.filename}")
        self.mtime = mtime or datetime.now()

    def clone(self) -> Todo:
        """
        Returns a clone of this todo

        Returns a copy of this todo, which is almost identical, except that is
        has a different UUID and filename.
        """
        todo = Todo(new=True, list=self.list)

        fields = [
            *Todo.STRING_FIELDS,
            *Todo.INT_FIELDS,
            *Todo.LIST_FIELDS,
            *Todo.DATETIME_FIELDS,
        ]
        fields.remove("uid")

        for field in fields:
            setattr(todo, field, getattr(self, field))

        return todo

    STRING_FIELDS = (
        "description",
        "location",
        "status",
        "summary",
        "uid",
        "rrule",
    )
    INT_FIELDS = (
        "percent_complete",
        "priority",
        "sequence",
    )
    LIST_FIELDS = ("categories",)
    DATETIME_FIELDS = (
        "completed_at",
        "created_at",
        "dtstamp",
        "start",
        "due",
        "last_modified",
    )
    RRULE_FIELDS = ("rrule",)
    ALL_SUPPORTED_FIELDS = (
        DATETIME_FIELDS + INT_FIELDS + LIST_FIELDS + RRULE_FIELDS + STRING_FIELDS
    )

    VALID_STATUSES = (
        "CANCELLED",
        "COMPLETED",
        "IN-PROCESS",
        "NEEDS-ACTION",
    )

    def __setattr__(self, name: str, value: None | str | int | list) -> None:
        """Check type and avoid setting fields to None"""
        """when that is not a valid attribue."""

        v = value

        if name in Todo.RRULE_FIELDS:
            if value is None:
                v = ""
            else:
                assert isinstance(
                    value, str
                ), f"Got {type(value)} for {name} where str was expected"

        if name in Todo.STRING_FIELDS:
            if value is None:
                v = ""
            else:
                assert isinstance(
                    value, str
                ), f"Got {type(value)} for {name} where str was expected"

        if name in Todo.INT_FIELDS:
            if value is None:
                v = 0
            else:
                assert isinstance(
                    value, int
                ), f"Got {type(value)} for {name} where int was expected"

        if name in Todo.LIST_FIELDS:
            if value is None:
                v = []
            else:
                assert isinstance(
                    value, list
                ), f"Got {type(value)} for {name} where list was expected"

        return object.__setattr__(self, name, v)

    @property
    def is_completed(self) -> bool:
        return bool(self.completed_at) or self.status in ("CANCELLED", "COMPLETED")

    @property
    def is_recurring(self) -> bool:
        return bool(self.rrule)

    def _apply_recurrence_to_dt(self, dt: date | None) -> datetime | None:
        if not dt:
            return None

        recurrence = rrulestr(self.rrule, dtstart=dt)

        if isinstance(dt, date) and not isinstance(dt, datetime):
            dt = datetime.combine(dt, time.min)

        return recurrence.after(dt)

    def _create_next_instance(self) -> Todo:
        copy = self.clone()
        copy.due = self._apply_recurrence_to_dt(self.due)
        copy.start = self._apply_recurrence_to_dt(self.start)

        assert copy.uid != self.uid

        # TODO: Push copy's alarms.
        return copy

    def complete(self) -> None:
        """
        Immediately completes this todo

        Immediately marks this todo as completed, sets the percentage to 100%
        and the completed_at datetime to now.

        If this todo belongs to a series, newly created todo are added to the
        ``related`` list.
        """
        if self.is_recurring:
            related = self._create_next_instance()
            if related:
                self.rrule = None
                self.related.append(related)

        self.completed_at = datetime.now(tz=LOCAL_TIMEZONE)
        self.percent_complete = 100
        self.status = "COMPLETED"

    @cached_property
    def path(self) -> str:
        if not self.list:
            raise ValueError("A todo without a list does not have a path.")

        return os.path.join(self.list.path, self.filename)

    def cancel(self) -> None:
        self.status = "CANCELLED"


class VtodoWriter:
    """Writes a Todo as a VTODO file."""

    """Maps Todo field names to VTODO field names"""
    FIELD_MAP = {
        "summary": "summary",
        "priority": "priority",
        "sequence": "sequence",
        "uid": "uid",
        "categories": "categories",
        "completed_at": "completed",
        "description": "description",
        "dtstamp": "dtstamp",
        "start": "dtstart",
        "due": "due",
        "location": "location",
        "percent_complete": "percent-complete",
        "status": "status",
        "created_at": "created",
        "last_modified": "last-modified",
        "rrule": "rrule",
    }

    def __init__(self, todo: Todo) -> None:
        self.todo = todo

    def normalize_datetime(self, dt: date) -> date:
        """
        Eliminate several differences between dates, times and datetimes which
        are hindering comparison:

        - Convert everything to datetime
        - Add missing timezones
        - Cast to UTC

        Datetimes are cast to UTC because icalendar doesn't include the
        VTIMEZONE information upon serialization, and some clients have issues
        dealing with that.
        """
        if isinstance(dt, date) and not isinstance(dt, datetime):
            return dt

        if not dt.tzinfo:
            dt = dt.replace(tzinfo=LOCAL_TIMEZONE)

        return dt.astimezone(timezone.utc)

    def serialize(self, original: icalendar.Todo = None) -> icalendar.Todo:
        """Serialize a Todo into a VTODO."""
        if not original:
            original = icalendar.Todo()
        self.vtodo = original

        for source, target in self.FIELD_MAP.items():
            self.vtodo.pop(target)
            if getattr(self.todo, source):
                raw = getattr(self.todo, source)
                if source in Todo.RRULE_FIELDS:
                    value = icalendar.vRecur.from_ical(raw)
                elif source in Todo.DATETIME_FIELDS:
                    value = self.normalize_datetime(raw)
                elif source in Todo.LIST_FIELDS:
                    value = raw
                elif source in Todo.INT_FIELDS:
                    value = int(raw)
                elif source in Todo.STRING_FIELDS:
                    value = raw
                else:
                    raise Exception(f"Unknown field {source} serialized.")

                self.vtodo.pop(target)  # Removes if None, avoids dupes
                if value:
                    logger.debug("Setting field %s to %s.", target, value)
                    self.vtodo.add(target, value)

        return self.vtodo

    def _read(self, path: str) -> icalendar.Todo:
        with open(path, "rb") as f:
            cal = f.read()
            data = icalendar.Calendar.from_ical(cal)
            for component in data.walk("VTODO"):
                return component
            return None

    def write(self) -> icalendar.Todo:
        if os.path.exists(self.todo.path):
            self._write_existing(self.todo.path)
        else:
            self._write_new(self.todo.path)

        return self.vtodo

    def _write_existing(self, path: str) -> None:
        original = self._read(path)
        vtodo = self.serialize(original)

        with open(path, "rb") as f:
            cal = icalendar.Calendar.from_ical(f.read())
            for index, component in enumerate(cal.subcomponents):
                if component.get("uid", None) == self.todo.uid:
                    cal.subcomponents[index] = vtodo

        with AtomicWriter(path, "wb", overwrite=True).open() as f:
            f.write(cal.to_ical())

    def _write_new(self, path: str) -> icalendar.Todo:
        vtodo = self.serialize()

        c = icalendar.Calendar()
        c.add_component(vtodo)

        with AtomicWriter(path, "wb").open() as f:
            c.add("prodid", "nl.whynothugo.todoman")
            c.add("version", "2.0")
            f.write(c.to_ical())

        return vtodo


class Cache:
    """
    Caches Todos for faster read and simpler querying interface

    The Cache class persists relevant[1] fields into an SQL database, which is
    only updated if the actual file has been modified. This greatly increases
    load times, but, more importantly, provides a simpler interface for
    filtering/querying/sorting.

    [1]: Relevant fields are those we show when listing todos, or those which
    may be used for filtering/sorting.
    """

    SCHEMA_VERSION = 10

    def __init__(self, path: str) -> None:
        self.cache_path = str(path)
        os.makedirs(os.path.dirname(self.cache_path), exist_ok=True)

        self._conn = sqlite3.connect(self.cache_path)
        self._conn.row_factory = sqlite3.Row
        self._conn.execute("PRAGMA foreign_keys = ON")

        self.create_tables()

    def save_to_disk(self) -> None:
        self._conn.commit()

    def is_latest_version(self) -> bool:
        """Checks if the cache DB schema is the latest version."""
        try:
            return self._conn.execute(
                "SELECT version FROM meta WHERE version = ?",
                (Cache.SCHEMA_VERSION,),
            ).fetchone()
        except sqlite3.OperationalError:
            return False

    def drop_tables(self) -> None:
        self._conn.executescript(
            """
            DROP TABLE IF EXISTS todos;
            DROP TABLE IF EXISTS lists;
            DROP TABLE IF EXISTS files;
            DROP TABLE IF EXISTS categories;
        """
        )

    def create_tables(self) -> None:
        if self.is_latest_version():
            return

        self.drop_tables()

        self._conn.execute('CREATE TABLE IF NOT EXISTS meta ("version" INT)')

        self._conn.execute(
            "INSERT INTO meta (version) VALUES (?)",
            (Cache.SCHEMA_VERSION,),
        )

        self._conn.execute(
            """
            CREATE TABLE IF NOT EXISTS lists (
                "name" TEXT PRIMARY KEY,
                "path" TEXT,
                "colour" TEXT,
                "mtime" INTEGER,

                CONSTRAINT path_unique UNIQUE (path)
            );
        """
        )

        self._conn.execute(
            """
            CREATE TABLE IF NOT EXISTS files (
                "path" TEXT PRIMARY KEY,
                "list_name" TEXT,
                "mtime" INTEGER,

                CONSTRAINT path_unique UNIQUE (path),
                FOREIGN KEY(list_name) REFERENCES lists(name) ON DELETE CASCADE
            );
        """
        )

        self._conn.execute(
            """
            CREATE TABLE IF NOT EXISTS categories (
                "id" INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
                "todos_id" INTEGER NOT NULL,
                "category" TEXT,
                "category_upper" TEXT,

                CONSTRAINT category_unique UNIQUE (todos_id,category),
                FOREIGN KEY(todos_id) REFERENCES todos(id) ON DELETE CASCADE
            );
        """
        )

        self._conn.execute(
            """
            CREATE TABLE IF NOT EXISTS todos (
                "file_path" TEXT,

                "id" INTEGER PRIMARY KEY,
                "uid" TEXT,
                "summary" TEXT,
                "due" INTEGER,
                "due_dt" INTEGER,
                "start" INTEGER,
                "start_dt" INTEGER,
                "priority" INTEGER,
                "created_at" INTEGER,
                "completed_at" INTEGER,
                "percent_complete" INTEGER,
                "dtstamp" INTEGER,
                "status" TEXT,
                "description" TEXT,
                "location" TEXT,
                "sequence" INTEGER,
                "last_modified" INTEGER,
                "rrule" TEXT,

                FOREIGN KEY(file_path) REFERENCES files(path) ON DELETE CASCADE
            );
        """
        )

    def clear(self) -> None:
        self._conn.close()
        os.remove(self.cache_path)

    def add_list(self, name: str, path: str, colour: str | None, mtime: int) -> str:
        """
        Inserts a new list into the cache.

        Returns the id of the newly inserted list.
        """

        result = self._conn.execute(
            "SELECT name FROM lists WHERE path = ?",
            (path,),
        ).fetchone()

        if result:
            return result["name"]

        try:
            self._conn.execute(
                """
                INSERT INTO lists (
                    name,
                    path,
                    colour,
                    mtime
                ) VALUES (?, ?, ?, ?)
                """,
                (
                    name,
                    path,
                    colour,
                    mtime,
                ),
            )
        except sqlite3.IntegrityError as e:
            raise exceptions.AlreadyExistsError("list", name) from e

        return self.add_list(name, path, colour, mtime)

    def add_file(self, list_name: str, path: str, mtime: int) -> None:
        try:
            self._conn.execute(
                """
                INSERT INTO files (
                    list_name,
                    path,
                    mtime
                ) VALUES (?, ?, ?);
                """,
                (
                    list_name,
                    path,
                    mtime,
                ),
            )
        except sqlite3.IntegrityError as e:
            raise exceptions.AlreadyExistsError("file", list_name) from e

    def add_category(self, todos_id: int, category: str) -> None:
        try:
            self._conn.execute(
                """
               INSERT INTO categories (
                   todos_id,
                   category,
                   category_upper
               ) VALUES (?, ?, ?);
               """,
                (todos_id, category, category.upper()),
            )
        except sqlite3.IntegrityError as e:
            raise exceptions.AlreadyExistsError("category", category) from e

    def _serialize_datetime(
        self,
        todo: icalendar.Todo,
        field: str,
    ) -> tuple[int | None, bool | None]:
        """
        Serialize a todo field in two value, the first one is the corresponding
        timestamp, the second one is a boolean indicating if the serialized
        value is a date or a datetime.

        :param icalendar.Todo todo: An icalendar component object
        :param str field: The name of the field to serialize
        """
        dt = todo.decoded(field, None)
        if not dt:
            return None, None

        is_date = isinstance(dt, date) and not isinstance(dt, datetime)
        if is_date:
            dt = datetime(dt.year, dt.month, dt.day)

        if not dt.tzinfo:
            dt = dt.replace(tzinfo=LOCAL_TIMEZONE)
        return dt.timestamp(), is_date

    def _serialize_rrule(self, todo: icalendar.Todo, field: str) -> str | None:
        rrule = todo.get(field)
        if not rrule:
            return None

        return rrule.to_ical().decode()

    def add_vtodo(
        self,
        todo: icalendar.Todo,
        file_path: str,
        id: int | None = None,
    ) -> int:
        """Adds a todo into the cache.

        :param todo: The icalendar component object on which
        """

        sql = """
            INSERT INTO todos (
                {}
                file_path,
                uid,
                summary,
                due,
                due_dt,
                start,
                start_dt,
                priority,
                created_at,
                completed_at,
                percent_complete,
                dtstamp,
                status,
                description,
                location,
                sequence,
                last_modified,
                rrule
            ) VALUES ({}?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?,
                ?)
            """

        due, due_dt = self._serialize_datetime(todo, "due")
        start, start_dt = self._serialize_datetime(todo, "dtstart")

        if start and due:
            start = None if start >= due else start

        params = [
            file_path,
            todo.get("uid"),
            todo.get("summary"),
            due,
            due_dt,
            start,
            start_dt,
            todo.get("priority", 0) or None,
            self._serialize_datetime(todo, "created")[0],
            self._serialize_datetime(todo, "completed")[0],
            todo.get("percent-complete", None),
            self._serialize_datetime(todo, "dtstamp")[0],
            todo.get("status", "NEEDS-ACTION"),
            todo.get("description", None),
            todo.get("location", None),
            todo.get("sequence", 1),
            self._serialize_datetime(todo, "last-modified")[0],
            self._serialize_rrule(todo, "rrule"),
        ]

        if id:
            params = [id, *params]
            sql = sql.format("id,\n", "?, ")
        else:
            sql = sql.format("", "")

        cursor = self._conn.cursor()
        try:
            cursor.execute(sql, params)
            rv = cursor.lastrowid
            assert rv is not None
        finally:
            cursor.close()

        if todo.get("categories"):
            for category in todo.get("categories").cats:
                self.add_category(rv, category)

        return rv

    def todos(
        self,
        *,
        lists: Iterable = (),
        categories: str | None = None,
        priority: str | None = None,
        location: str = "",
        grep: str = "",
        sort: Iterable = (),
        reverse: bool = True,
        due: int | None = None,
        start: tuple[bool, datetime] | None = None,
        startable: bool = False,
        status: str = "NEEDS-ACTION,IN-PROCESS",
    ) -> Iterator[Todo]:
        """
        Returns filtered cached todos, in a specified order.

        If no order is specified, todos are sorted by the following fields::

            completed_at
            -priority
            due
            -created_at

        :param lists: Only return todos for these lists.
        :param location: Only return todos with a location containing this string.
        :param categories: Only return todos with a category containing this string.
        :param grep: Filter common fields with this substring.
        :param sort: Order returned todos by these fields. Field names
            with a ``-`` prepended will be used to sort in reverse order.
        :param reverse: Reverse the order of the todos after sorting.
        :param due: Return only todos due within ``due`` hours.
        :param priority: Only return todos with priority at least as high as specified.
        :param start: Return only todos before/after ``start`` date
        :param status: Return only todos with any of the given statuses.
        :return: A sorted, filtered list of todos.
        """
        extra_where = []
        params: list = []

        if "ANY" not in status:
            statuses = status.split(",")
            extra_where.append(
                "AND (status IN ({}) OR status IS NULL)".format(
                    ", ".join(["?"] * len(statuses))
                )
            )
            params.extend(s.upper() for s in statuses)

        if lists:
            lists = [
                list_.name if isinstance(list_, TodoList) else list_ for list_ in lists
            ]
            q = ", ".join(["?"] * len(lists))
            extra_where.append(f"AND files.list_name IN ({q})")
            params.extend(lists)
        if categories:
            category_slots = ", ".join(["?"] * len(categories))
            extra_where.append(f"AND categories.category_upper IN ({category_slots})")
            params = params + [category.upper() for category in categories]
        if priority:
            extra_where.append("AND PRIORITY > 0 AND PRIORITY <= ?")
            params.append(f"{priority}")
        if location:
            extra_where.append("AND location LIKE ?")
            params.append(f"%{location}%")
        if grep:
            # # requires sqlite with pcre, which won't be available everywhere:
            # extra_where.append('AND summary REGEXP ?')
            # params.append(grep)
            extra_where.append("AND summary LIKE ?")
            params.append(f"%{grep}%")
        if due:
            max_due = (datetime.now() + timedelta(hours=due)).timestamp()
            extra_where.append("AND due IS NOT NULL AND due < ?")
            params.append(max_due)
        if start:
            is_before, dt = start
            timestamp = dt.timestamp()
            if is_before:
                extra_where.append("AND start <= ?")
                params.append(timestamp)
            else:
                extra_where.append("AND start >= ?")
                params.append(timestamp)
        if startable:
            extra_where.append("AND (start IS NULL OR start <= ?)")
            params.append(datetime.now().timestamp())
        if sort:
            order_items = []
            for s in sort:
                if s.startswith("-"):
                    order_items.append(f" {s[1:]} ASC")
                else:
                    order_items.append(f" {s} DESC")
            order = ",".join(order_items)
        else:
            order = """
                completed_at DESC,
                priority IS NOT NULL, priority DESC,
                due IS NOT NULL, due DESC,
                created_at ASC
            """

        if not reverse:
            # Note the change in case to avoid swapping all of them. sqlite
            # doesn't care about casing anyway.
            order = order.replace(" DESC", " asc").replace(" ASC", " desc")

        query = """
        SELECT DISTINCT todos.*, files.list_name, files.path,
          group_concat(category) AS categories
        FROM todos, files
        LEFT JOIN categories
        ON categories.todos_id = todos.id
        WHERE todos.file_path = files.path {}
        GROUP BY uid
        ORDER BY {}
        """.format(
            " ".join(extra_where),
            order,
        )

        logger.debug(query)
        logger.debug(params)

        result = self._conn.execute(query, params)

        seen_paths = set()
        warned_paths = set()

        for row in result:
            todo = self._todo_from_db(row)
            path = row["path"]

            if path in seen_paths and path not in warned_paths:
                logger.warning(
                    "Todo is in read-only mode because there are multiple todos in %s",
                    path,
                )
                warned_paths.add(path)
            seen_paths.add(path)
            yield todo

    def _datetime_from_db(self, dt: float) -> datetime | None:
        if dt:
            return datetime.fromtimestamp(dt, LOCAL_TIMEZONE)
        return None

    def _date_from_db(self, dt: float, is_date: bool = False) -> date | None:
        """Deserialise a date (possible datetime)."""
        if not dt:
            return None

        if is_date:
            return datetime.fromtimestamp(dt, LOCAL_TIMEZONE).date()
        return datetime.fromtimestamp(dt, LOCAL_TIMEZONE)

    def _categories_from_db(self, categories: str) -> list[str]:
        if categories:
            return categories.split(",")

        return []

    def _todo_from_db(self, row: dict) -> Todo:
        todo = Todo()
        todo.id = row["id"]
        todo.uid = row["uid"]
        todo.summary = row["summary"]
        todo.due = self._date_from_db(row["due"], row["due_dt"])
        todo.start = self._date_from_db(row["start"], row["start_dt"])
        todo.categories = self._categories_from_db(row["categories"])
        todo.priority = row["priority"]
        todo.created_at = self._datetime_from_db(row["created_at"])
        todo.completed_at = self._datetime_from_db(row["completed_at"])
        todo.dtstamp = self._datetime_from_db(row["dtstamp"])
        todo.percent_complete = row["percent_complete"]
        todo.status = row["status"]
        todo.description = row["description"]
        todo.location = row["location"]

        logger.debug("todo.categories: %s\n", todo.categories)
        todo.sequence = row["sequence"]
        todo.last_modified = row["last_modified"]
        todo.list = self.lists_map[row["list_name"]]
        todo.filename = os.path.basename(row["path"])
        todo.rrule = row["rrule"]
        return todo

    def lists(self) -> Iterator[TodoList]:
        result = self._conn.execute("SELECT * FROM lists")
        for row in result:
            yield TodoList(
                name=row["name"],
                path=row["path"],
                colour=row["colour"],
            )

    @cached_property
    def lists_map(self) -> dict[str, TodoList]:
        return {list_.name: list_ for list_ in self.lists()}

    def expire_lists(self, paths: dict[str, int]) -> None:
        results = self._conn.execute("SELECT path, name, mtime from lists")
        for result in results:
            if result["path"] not in paths:
                self.delete_list(result["name"])
            else:
                mtime = paths.get(result["path"])
                if mtime and mtime > result["mtime"]:
                    self.delete_list(result["name"])

    def delete_list(self, name: str) -> None:
        self._conn.execute("DELETE FROM lists WHERE lists.name = ?", (name,))

    def todo(self, id: int, read_only: bool = False) -> Todo:
        # XXX: DON'T USE READ_ONLY
        result = self._conn.execute(
            """
            SELECT todos.*, files.list_name, files.path,
              group_concat(category) AS categories
            FROM todos, files
            LEFT JOIN categories
            ON categories.todos_id = todos.id
            WHERE files.path = todos.file_path
              AND todos.id = ?
            GROUP BY uid
        """,
            (id,),
        ).fetchone()

        if not result:
            raise exceptions.NoSuchTodoError(id)

        if not read_only:
            count = self._conn.execute(
                """
                SELECT count(id) AS c
                  FROM files, todos
                 WHERE todos.file_path = files.path
                   AND path=?
            """,
                (result["path"],),
            ).fetchone()
            if count["c"] > 1:
                raise exceptions.ReadOnlyTodoError(result["path"])

        return self._todo_from_db(result)

    def expire_files(self, paths_to_mtime: dict[str, int]) -> None:
        """Remove stale cache entries based on the given fresh data."""
        result = self._conn.execute("SELECT path, mtime FROM files")
        for row in result:
            path, mtime = row["path"], row["mtime"]
            if paths_to_mtime.get(path) != mtime:
                self.expire_file(path)

    def expire_file(self, path: str) -> None:
        self._conn.execute("DELETE FROM files WHERE path = ?", (path,))


class TodoList:
    def __init__(self, name: str, path: str, colour: str | None = None) -> None:
        self.path = path
        self.name = name
        self.colour = colour

    @staticmethod
    def colour_for_path(path: str) -> str | None:
        try:
            with open(os.path.join(path, "color")) as f:
                return f.read().strip()
        except OSError:
            logger.debug("No colour for list %s", path)

        return None

    @staticmethod
    def name_for_path(path: str) -> str:
        try:
            with open(os.path.join(path, "displayname")) as f:
                return f.read().strip()
        except OSError:
            return split(normpath(path))[1]

    @staticmethod
    def mtime_for_path(path: str) -> int:
        colour_file = os.path.join(path, "color")
        display_file = os.path.join(path, "displayname")

        mtimes = []
        if os.path.exists(colour_file):
            mtimes.append(_getmtime(colour_file))
        if os.path.exists(display_file):
            mtimes.append(_getmtime(display_file))

        if mtimes:
            return max(mtimes)
        return 0

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TodoList):
            return self.name == other.name
        return object.__eq__(self, other)

    def __str__(self) -> str:
        return self.name


class Database:
    """
    This class is essentially a wrapper around all the lists (which in turn,
    contain all the todos).

    Caching in abstracted inside this class, and is transparent to outside
    classes.
    """

    def __init__(self, paths: Iterable[str], cache_path: str) -> None:
        self.cache = Cache(cache_path)
        self.paths = [str(path) for path in paths]
        self.update_cache()

    def update_cache(self) -> None:
        paths = {path: TodoList.mtime_for_path(path) for path in self.paths}
        self.cache.expire_lists(paths)

        paths_to_mtime = {}
        paths_to_list_name = {}

        for path in self.paths:
            list_name = self.cache.add_list(
                TodoList.name_for_path(path),
                path,
                TodoList.colour_for_path(path),
                paths[path],
            )
            for entry in os.listdir(path):
                if not entry.endswith(".ics"):
                    continue
                entry_path = os.path.join(path, entry)
                mtime = _getmtime(entry_path)
                paths_to_mtime[entry_path] = mtime
                paths_to_list_name[entry_path] = list_name

        self.cache.expire_files(paths_to_mtime)

        for entry_path, mtime in paths_to_mtime.items():
            list_name = paths_to_list_name[entry_path]

            try:
                self.cache.add_file(list_name, entry_path, mtime)
            except exceptions.AlreadyExistsError:
                logger.debug("File already in cache: %s", entry_path)
                continue

            try:
                with open(entry_path, "rb") as f:
                    cal = icalendar.Calendar.from_ical(f.read())
                    for component in cal.walk("VTODO"):
                        self.cache.add_vtodo(component, entry_path)
            except Exception:
                logger.exception("Failed to read entry %s.", entry_path)

        self.cache.save_to_disk()

    def todos(self, **kwargs) -> Iterator[Todo]:
        return self.cache.todos(**kwargs)

    def todo(self, id: int, **kwargs) -> Todo:
        return self.cache.todo(id, **kwargs)

    def lists(self) -> Iterator[TodoList]:
        return self.cache.lists()

    def move(self, todo: Todo, new_list: TodoList, from_list: TodoList) -> None:
        orig_path = os.path.join(from_list.path, todo.filename)
        dest_path = os.path.join(new_list.path, todo.filename)

        os.rename(orig_path, dest_path)

    def delete(self, todo: Todo) -> None:
        if not todo.list:
            raise ValueError("Cannot delete Todo without a list.")

        path = os.path.join(todo.list.path, todo.filename)
        os.remove(path)

    def flush(self) -> Iterable[Todo]:
        for todo in self.todos(status=["ANY"]):
            if todo.is_completed:
                yield todo
                self.delete(todo)

        self.cache.clear()
        self.cache = Cache(self.cache.cache_path)

    def save(self, todo: Todo) -> None:
        if not todo.list:
            raise ValueError("Cannot save Todo without a list.")

        for related in todo.related:
            self.save(related)

        todo.sequence += 1
        todo.last_modified = datetime.now(LOCAL_TIMEZONE)

        vtodo = VtodoWriter(todo).write()

        self.cache.expire_file(todo.path)
        mtime = _getmtime(todo.path)

        self.cache.add_file(todo.list.name, todo.path, mtime)
        todo.id = self.cache.add_vtodo(vtodo, todo.path, todo.id)
        self.cache.save_to_disk()


def _getmtime(path: str) -> int:
    return os.stat(path).st_mtime_ns



from __future__ import annotations

import contextlib
import json
from abc import ABC
from abc import abstractmethod
from datetime import date
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from datetime import tzinfo
from time import mktime
from typing import Iterable

import click
import humanize
import parsedatetime
from dateutil.tz import tzlocal

from todoman.model import Todo
from todoman.model import TodoList


def rgb_to_ansi(colour: str | None) -> str | None:
    """
    Convert a string containing an RGB colour to ANSI escapes
    """
    if not colour or not colour.startswith("#"):
        return None

    r, g, b = colour[1:3], colour[3:5], colour[5:7]

    if not len(r) == len(g) == len(b) == 2:
        return None

    return f"\33[38;2;{int(r, 16)!s};{int(g, 16)!s};{int(b, 16)!s}m"


class Formatter(ABC):
    @abstractmethod
    def __init__(
        self,
        date_format: str = "%Y-%m-%d",
        time_format: str = "%H:%M",
        dt_separator: str = " ",
    ) -> None:
        """Create a new formatter instance."""

    @abstractmethod
    def compact(self, todo: Todo) -> str:
        """Render a compact todo (usually in a single line)"""

    @abstractmethod
    def compact_multiple(self, todos: Iterable[Todo], hide_list: bool = False) -> str:
        """Same as compact() but for multiple todos."""

    @abstractmethod
    def simple_action(self, action: str, todo: Todo) -> str:
        """Render an action related to a todo (e.g.: compelete, undo, etc)."""

    @abstractmethod
    def parse_priority(self, priority: str | None) -> int | None:
        """Parse a priority"""

    @abstractmethod
    def detailed(self, todo: Todo) -> str:
        """Returns a detailed representation of a task."""

    @abstractmethod
    def format_datetime(self, value: date | None) -> str | int | None:
        """Format an optional datetime."""

    @abstractmethod
    def parse_datetime(self, value: str | None) -> date | None:
        """Parse an optional datetime."""

    @abstractmethod
    def format_database(self, database: TodoList) -> str:
        """Format the name of a single database."""

    @abstractmethod
    def parse_categories(self, categories: str) -> list[str]:
        """Parse multiple categories."""

    @abstractmethod
    def format_categories(self, categories: Iterable[str]) -> str:
        """Format multiple categories."""

    @abstractmethod
    def format_priority(self, priority: int | None) -> str:
        """Format a todo priority"""


class DefaultFormatter(Formatter):
    def __init__(
        self,
        date_format: str = "%Y-%m-%d",
        time_format: str = "%H:%M",
        dt_separator: str = " ",
        tz_override: tzinfo | None = None,
    ) -> None:
        self.date_format = date_format
        self.time_format = time_format
        self.dt_separator = dt_separator
        self.datetime_format = dt_separator.join(
            filter(bool, (date_format, time_format))
        )

        self.tz = tz_override or tzlocal()
        self.now = datetime.now().replace(tzinfo=self.tz)

        self._parsedatetime_calendar = parsedatetime.Calendar(
            version=parsedatetime.VERSION_CONTEXT_STYLE,
        )

    def simple_action(self, action: str, todo: Todo) -> str:
        return f'{action} "{todo.summary}"'

    def compact(self, todo: Todo) -> str:
        return self.compact_multiple([todo])

    def compact_multiple(self, todos: Iterable[Todo], hide_list: bool = False) -> str:
        # TODO: format lines fuidly and drop the table
        # it can end up being more readable when too many columns are empty.
        # show dates that are in the future in yellow (in 24hs) or grey (future)
        table = []
        for todo in todos:
            completed = "X" if todo.is_completed else " "
            percent = todo.percent_complete or ""
            if percent:
                percent = f" ({percent}%)"

            if todo.categories:
                categories = " [" + ", ".join(todo.categories) + "]"
            else:
                categories = ""

            priority = click.style(
                self.format_priority_compact(todo.priority),
                fg="magenta",
            )

            due = self.format_datetime(todo.due) or "(no due date)"
            due_colour = self._due_colour(todo)
            if due_colour:
                due = click.style(str(due), fg=due_colour)

            recurring = "⟳" if todo.is_recurring else ""

            if hide_list:
                summary = f"{todo.summary} {percent}"
            else:
                if not todo.list:
                    raise ValueError("Cannot format todo without a list")

                summary = f"{todo.summary} {self.format_database(todo.list)}{percent}"

            # TODO: add spaces on the left based on max todos"

            # FIXME: double space when no priority
            # split into parts to satisfy linter line too long
            table.append(
                f"[{completed}] {todo.id} {priority} {due} "
                f"{recurring}{summary}{categories}"
            )

        return "\n".join(table)

    def _due_colour(self, todo: Todo) -> str:
        now = self.now if isinstance(todo.due, datetime) else self.now.date()
        if todo.due:
            if todo.due <= now and not todo.is_completed:
                return "red"
            if todo.due >= now + timedelta(hours=24):
                return "white"
            if todo.due >= now:
                return "yellow"

        return "white"

    def _format_multiline(self, title: str, value: str) -> str:
        formatted_title = click.style(title, fg="white")

        if value.strip().count("\n") == 0:
            return f"\n\n{formatted_title}: {value}"
        return f"\n\n{formatted_title}:\n{value}"

    def detailed(self, todo: Todo) -> str:
        extra_lines = []
        if todo.description:
            extra_lines.append(self._format_multiline("Description", todo.description))

        if todo.location:
            extra_lines.append(self._format_multiline("Location", todo.location))

        return f"{self.compact(todo)}{''.join(extra_lines)}"

    # FIXME: cannot return `int`, but porcelain subclasses this (it shouldn't)
    def format_datetime(self, dt: date | None) -> str | int | None:
        if not dt:
            return ""
        if isinstance(dt, datetime):
            return dt.strftime(self.datetime_format)
        if isinstance(dt, date):
            return dt.strftime(self.date_format)
        return None

    def format_categories(self, categories: Iterable[str]) -> str:
        return ", ".join(categories)

    def parse_categories(self, categories: str) -> list[str]:
        # existing code assumes categories is list,
        # but click passes tuple
        return list(categories)

    def parse_priority(self, priority: str | None) -> int | None:
        if priority is None or priority == "":
            return None
        if priority == "low":
            return 9
        if priority == "medium":
            return 5
        if priority == "high":
            return 4
        if priority == "none":
            return 0
        raise ValueError("Priority has to be one of low, medium, high or none")

    def format_priority(self, priority: int | None) -> str:
        if not priority:
            return "none"
        if 1 <= priority <= 4:
            return "high"
        if priority == 5:
            return "medium"
        if 6 <= priority <= 9:
            return "low"

        raise ValueError("priority is an invalid value")

    def format_priority_compact(self, priority: int | None) -> str:
        if not priority:
            return ""
        if 1 <= priority <= 4:
            return "!!!"
        if priority == 5:
            return "!!"
        if 6 <= priority <= 9:
            return "!"

        raise ValueError("priority is an invalid value")

    def parse_datetime(self, dt: str | None) -> date | None:
        if not dt:
            return None

        rv = self._parse_datetime_naive(dt)
        return rv.replace(tzinfo=self.tz) if isinstance(rv, datetime) else rv

    def _parse_datetime_naive(self, dt: str) -> date:
        """Parse dt and returns a naive datetime or a date"""
        with contextlib.suppress(ValueError):
            return datetime.strptime(dt, self.datetime_format)

        with contextlib.suppress(ValueError):
            return datetime.strptime(dt, self.date_format).date()

        with contextlib.suppress(ValueError):
            return datetime.combine(
                self.now.date(), datetime.strptime(dt, self.time_format).time()
            )

        rv, pd_ctx = self._parsedatetime_calendar.parse(dt)
        if not pd_ctx.hasDateOrTime:
            raise ValueError(f"Time description not recognized: {dt}")
        return datetime.fromtimestamp(mktime(rv))

    def format_database(self, database: TodoList) -> str:
        return "{}@{}".format(
            rgb_to_ansi(database.colour) or "", click.style(database.name)
        )


class HumanizedFormatter(DefaultFormatter):
    def format_datetime(self, dt: date | None) -> str:
        if not dt:
            return ""

        if isinstance(dt, datetime):
            rv = humanize.naturaltime(self.now - dt)
            if " from now" in rv:
                rv = f"in {rv[:-9]}"
        elif isinstance(dt, date):
            rv = humanize.naturaldate(dt)

        return rv


class PorcelainFormatter(DefaultFormatter):
    def _todo_as_dict(self, todo: Todo) -> dict:
        return {
            "completed": todo.is_completed,
            "start": self.format_datetime(todo.start),
            "due": self.format_datetime(todo.due),
            "id": todo.id,
            "list": todo.list.name if todo.list else None,
            "percent": todo.percent_complete,
            "summary": todo.summary,
            "categories": todo.categories,
            "priority": todo.priority,
            "location": todo.location,
            "description": todo.description,
            "completed_at": self.format_datetime(todo.completed_at),
        }

    def compact(self, todo: Todo) -> str:
        return json.dumps(self._todo_as_dict(todo), indent=4, sort_keys=True)

    def compact_multiple(self, todos: Iterable[Todo], hide_list: bool = False) -> str:
        data = [self._todo_as_dict(todo) for todo in todos]
        return json.dumps(data, indent=4, sort_keys=True)

    def simple_action(self, action: str, todo: Todo) -> str:
        return self.compact(todo)

    def parse_priority(self, priority: str | None) -> int | None:
        if priority is None:
            return None
        try:
            if int(priority) in range(10):
                return int(priority)
            raise ValueError("Priority has to be in the range 0-9")
        except ValueError as e:
            raise click.BadParameter(str(e)) from None

    def detailed(self, todo: Todo) -> str:
        return self.compact(todo)

    def format_datetime(self, value: date | None) -> int | None:
        if value:
            if not isinstance(value, datetime):
                dt = datetime.fromordinal(value.toordinal())
            else:
                dt = value
            return int(dt.timestamp())
        return None

    def parse_datetime(self, value: str | float | None) -> datetime | None:
        if value:
            return datetime.fromtimestamp(float(value), tz=timezone.utc)
        return None



from __future__ import annotations

import functools
import glob
import locale
import sys
from contextlib import contextmanager
from datetime import date
from datetime import timedelta
from os.path import isdir
from typing import Callable
from typing import Iterator
from typing import Literal
from typing import ParamSpec
from typing import TypeVar

import click
import click_log

from todoman import exceptions
from todoman import formatters
from todoman.configuration import ConfigurationError
from todoman.configuration import load_config
from todoman.interactive import TodoEditor
from todoman.model import Database
from todoman.model import Todo
from todoman.model import TodoList
from todoman.model import cached_property

click_log.basic_config()


@contextmanager
def handle_error() -> Iterator[None]:
    try:
        yield
    except exceptions.TodomanError as e:
        click.echo(e)
        sys.exit(e.EXIT_CODE)


_T = TypeVar("_T")
_P = ParamSpec("_P")


def catch_errors(f: Callable[_P, _T]) -> Callable[_P, _T]:
    @functools.wraps(f)
    def wrapper(*a, **kw) -> _T:
        with handle_error():
            return f(*a, **kw)

    return wrapper


TODO_ID_MIN = 1
with_id_arg = click.argument("id", type=click.IntRange(min=TODO_ID_MIN))


def _validate_lists_param(
    ctx: click.Context,
    param: click.Parameter,
    lists: list[str],
) -> list[TodoList]:
    return [_validate_list_param(ctx, param, name=list_) for list_ in lists]


def _validate_list_param(
    click_ctx: click.Context,
    param: click.Parameter,
    name: str,
) -> TodoList:
    ctx = click_ctx.find_object(AppContext)
    assert ctx, "AppContext must be defined"
    if name is None:
        if ctx.config["default_list"]:
            name = ctx.config["default_list"]
        else:
            raise click.BadParameter("You must set `default_list` or use -l.")
    lists = {list_.name: list_ for list_ in ctx.db.lists()}
    fuzzy_matches = [
        list_ for list_ in lists.values() if list_.name.lower() == name.lower()
    ]

    if len(fuzzy_matches) == 1:
        return fuzzy_matches[0]

    # case-insensitive matching collides or does not find a result,
    # use exact matching
    if name in lists:
        return lists[name]
    raise click.BadParameter(
        "{}. Available lists are: {}".format(
            name, ", ".join(list_.name for list_ in lists.values())
        )
    )


def _validate_date_param(
    click_ctx: click.Context,
    param: click.Parameter,
    val: str,
) -> date | None:
    ctx = click_ctx.find_object(AppContext)
    assert ctx, "AppContext must be defined"
    try:
        return ctx.formatter.parse_datetime(val)
    except ValueError as e:
        raise click.BadParameter(str(e)) from None


def _validate_categories_param(
    click_ctx: click.Context,
    param: click.Parameter,
    val: str,
) -> list[str]:
    ctx = click_ctx.find_object(AppContext)
    assert ctx, "AppContext must be defined"
    return ctx.formatter.parse_categories(val)


def _validate_priority_param(
    click_ctx: click.Context,
    param: click.Parameter,
    val: str,
) -> int | None:
    ctx = click_ctx.find_object(AppContext)
    assert ctx, "AppContext must be defined"
    try:
        return ctx.formatter.parse_priority(val)
    except ValueError as e:
        raise click.BadParameter(str(e)) from None


def _validate_start_date_param(
    click_ctx: click.Context,
    param: click.Parameter,
    val: str,
) -> tuple[bool, date] | None:
    ctx = click_ctx.find_object(AppContext)
    assert ctx, "AppContext must be defined"
    if not val:
        return None

    if len(val) != 2 or val[0] not in ["before", "after"]:
        raise click.BadParameter("Format should be '[before|after] [DATE]'")

    is_before = val[0] == "before"

    try:
        dt = ctx.formatter.parse_datetime(val[1])
        if dt is None:
            return None
        return is_before, dt
    except ValueError as e:
        raise click.BadParameter(str(e)) from None


def _validate_startable_param(
    click_ctx: click.Context,
    param: click.Parameter,
    val: bool,
) -> bool:
    ctx = click_ctx.find_object(AppContext)
    assert ctx, "AppContext must be defined"
    return val or ctx.config["startable"]


def _validate_todos(
    click_ctx: click.Context,
    param: click.Parameter,
    val: list[int],
) -> list[Todo]:
    ctx = click_ctx.find_object(AppContext)
    assert ctx, "AppContext must be defined"
    with handle_error():
        return [ctx.db.todo(int(id)) for id in val]


def _sort_callback(ctx: click.Context, param: click.Parameter, val: str) -> list[str]:
    fields = val.split(",") if val else []
    for field in fields:
        if field.startswith("-"):
            field = field[1:]

        if field not in Todo.ALL_SUPPORTED_FIELDS and field != "id":
            raise click.BadParameter(f"Unknown field '{field}'")

    return fields


def validate_status(ctx: AppContext, param: click.Parameter, val: str) -> str:
    statuses = val.upper().split(",")

    if "ANY" in statuses:
        return ",".join(Todo.VALID_STATUSES)

    for status in statuses:
        if status not in Todo.VALID_STATUSES:
            raise click.BadParameter(
                'Invalid status, "{}", statuses must be one of "{}", or "ANY"'.format(
                    status, ", ".join(Todo.VALID_STATUSES)
                )
            )

    return val


def _todo_property_options(command: Callable) -> Callable:
    click.option(
        "--category",
        "-c",
        multiple=True,
        default=(),
        callback=_validate_categories_param,
        help="Task categories. Can be used multiple times.",
    )(command)
    click.option(
        "--priority",
        default="",
        callback=_validate_priority_param,
        help="Priority for this task",
    )(command)
    click.option("--location", help="The location where this todo takes place.")(
        command
    )
    click.option(
        "--due",
        "-d",
        default="",
        callback=_validate_date_param,
        help=("Due date of the task, in the format specified in the configuration."),
    )(command)
    click.option(
        "--start",
        "-s",
        default="",
        callback=_validate_date_param,
        help="When the task starts.",
    )(command)

    @functools.wraps(command)
    def command_wrap(*a, **kw) -> click.Command:
        kw["todo_properties"] = {
            key: kw.pop(key) for key in ("due", "start", "location", "priority")
        }
        # longform is singular since user can pass it multiple times, but
        # in actuality it's plural, so manually changing for #cache.todos.
        kw["todo_properties"]["categories"] = kw.pop("category")

        return command(*a, **kw)

    return command_wrap


class AppContext:
    config: dict  # TODO: better typing
    db: Database
    formatter_class: type[formatters.Formatter]

    @cached_property
    def ui_formatter(self) -> formatters.Formatter:
        return formatters.DefaultFormatter(
            self.config["date_format"],
            self.config["time_format"],
            self.config["dt_separator"],
        )

    @cached_property
    def formatter(self) -> formatters.Formatter:
        return self.formatter_class(
            self.config["date_format"],
            self.config["time_format"],
            self.config["dt_separator"],
        )


pass_ctx = click.make_pass_decorator(AppContext)

_interactive_option = click.option(
    "--interactive",
    "-i",
    is_flag=True,
    default=None,
    help="Go into interactive mode before saving the task.",
)


@click.group(invoke_without_command=True)
@click_log.simple_verbosity_option()
@click.option(
    "--colour",
    "--color",
    "colour",
    default=None,
    type=click.Choice(["always", "auto", "never"]),
    help=(
        "By default todoman will disable colored output if stdout "
        "is not a TTY (value `auto`). Set to `never` to disable "
        "colored output entirely, or `always` to enable it "
        "regardless."
    ),
)
@click.option(
    "--porcelain",
    is_flag=True,
    help=(
        "Use a JSON format that will "
        "remain stable regardless of configuration or version."
    ),
)
@click.option(
    "--humanize",
    "-h",
    default=None,
    is_flag=True,
    help="Format all dates and times in a human friendly way",
)
@click.option(
    "--config",
    "-c",
    default=None,
    help="The config file to use.",
    envvar="TODOMAN_CONFIG",
    metavar="PATH",
)
@click.pass_context
@click.version_option(prog_name="todoman")
@catch_errors
def cli(
    click_ctx: click.Context,
    colour: Literal["always"] | Literal["auto"] | Literal["never"],
    porcelain: bool,
    humanize: bool,
    config: str,
) -> None:
    ctx = click_ctx.ensure_object(AppContext)
    try:
        ctx.config = load_config(config)
    except ConfigurationError as e:
        raise click.ClickException(e.args[0]) from None

    if porcelain and humanize:
        raise click.ClickException(
            "--porcelain and --humanize cannot be used at the same time."
        )

    if humanize is None:  # False means explicitly disabled
        humanize = ctx.config["humanize"]

    if porcelain:
        ctx.formatter_class = formatters.PorcelainFormatter
    elif humanize:
        ctx.formatter_class = formatters.HumanizedFormatter
    else:
        ctx.formatter_class = formatters.DefaultFormatter

    colour = colour or ctx.config["color"]
    if colour == "always":
        click_ctx.color = True
    elif colour == "never":
        click_ctx.color = False

    paths = [
        path
        for path in glob.iglob(ctx.config["path"])
        if isdir(path) and not path.endswith("__pycache__")
    ]
    if len(paths) == 0:
        raise exceptions.NoListsFoundError(ctx.config["path"])

    ctx.db = Database(paths, ctx.config["cache_path"])

    # Make python actually use LC_TIME, or the user's locale settings
    locale.setlocale(locale.LC_TIME, "")

    if not click_ctx.invoked_subcommand:
        invoke_command(
            click_ctx,
            ctx.config["default_command"],
        )


def invoke_command(click_ctx: click.Context, command: str) -> None:
    name, *raw_args = command.split(" ")
    if name not in cli.commands:
        raise click.ClickException("Invalid setting for [default_command]")
    parser = cli.commands[name].make_parser(click_ctx)
    opts, args, param_order = parser.parse_args(raw_args)
    for param in param_order:
        if param.name:
            opts[param.name] = param.handle_parse_result(click_ctx, opts, args)[0]
    click_ctx.invoke(cli.commands[name], *args, **opts)


@cli.command()
@click.pass_context
def repl(ctx: click.Context) -> None:
    """Start an interactive shell."""
    try:
        from click_repl import repl

        repl(ctx)
    except ImportError as e:
        click.echo(e)
        sys.exit(-1)


@cli.command()
@click.argument("summary", nargs=-1)
@click.option(
    "--list",
    "-l",
    callback=_validate_list_param,
    help="List in which the task will be saved.",
)
@click.option(
    "--read-description",
    "-r",
    is_flag=True,
    default=False,
    help="Read task description from stdin.",
)
@_todo_property_options
@_interactive_option
@pass_ctx
@catch_errors
def new(
    ctx: AppContext,
    summary: str,
    list: TodoList,
    todo_properties: dict,
    read_description: bool,
    interactive: bool,
) -> None:
    """
    Create a new task with SUMMARY.
    """

    todo = Todo(new=True, list=list)

    default_due = ctx.config["default_due"]
    if default_due:
        assert todo.created_at, "create_at must be defined for new Todo"
        todo.due = todo.created_at + timedelta(hours=default_due)

    default_priority = ctx.config["default_priority"]
    if default_priority is not None:
        todo.priority = default_priority

    for key, value in todo_properties.items():
        if value is not None:
            setattr(todo, key, value)
    todo.summary = " ".join(summary)

    if read_description:
        todo.description = "\n".join(sys.stdin)

    if interactive or (not summary and interactive is None):
        ui = TodoEditor(todo, ctx.db.lists(), ctx.ui_formatter)
        ui.edit()
        click.echo()  # work around lines going missing after urwid

    if not todo.summary:
        raise click.UsageError("No SUMMARY specified")

    ctx.db.save(todo)
    click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@click.option(
    "--raw",
    is_flag=True,
    help=(
        "Open the raw file for editing in $EDITOR.\n"
        "Only use this if you REALLY know what you're doing!"
    ),
)
@_todo_property_options
@_interactive_option
@with_id_arg
@catch_errors
def edit(
    ctx: AppContext,
    id: int,
    todo_properties: dict,
    interactive: bool,
    raw: bool,
) -> None:
    """
    Edit the task with id ID.
    """
    todo = ctx.db.todo(id)
    if raw:
        click.edit(filename=todo.path)
        return
    old_list = todo.list

    changes = False
    for key, value in todo_properties.items():
        if value is not None and value != []:
            changes = True
            setattr(todo, key, value)

    if interactive or (not changes and interactive is None):
        ui = TodoEditor(todo, ctx.db.lists(), ctx.ui_formatter)
        ui.edit()

    # This little dance avoids duplicates when changing the list:
    new_list = todo.list
    assert new_list, "New list cannot be None"
    assert old_list, "Original list cannot be None"
    todo.list = old_list
    ctx.db.save(todo)
    if old_list != new_list:
        ctx.db.move(todo, new_list=new_list, from_list=old_list)
    click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@with_id_arg
@catch_errors
def show(ctx: AppContext, id: int) -> None:
    """
    Show details about a task.
    """
    todo = ctx.db.todo(id, read_only=True)
    click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@with_id_arg
@catch_errors
def path(ctx: AppContext, id: int) -> None:
    """Print the path to a task's file."""
    todo = ctx.db.todo(id, read_only=True)
    click.echo(todo.path)


@cli.command()
@pass_ctx
@click.argument(
    "todos",
    nargs=-1,
    required=True,
    type=click.IntRange(0),
    callback=_validate_todos,
)
@catch_errors
def done(ctx: AppContext, todos: list[Todo]) -> None:
    """Mark one or more tasks as done."""
    for todo in todos:
        todo.complete()
        ctx.db.save(todo)
        click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@click.argument(
    "todos",
    nargs=-1,
    required=True,
    type=click.IntRange(0),
    callback=_validate_todos,
)
@catch_errors
def cancel(ctx: AppContext, todos: list[Todo]) -> None:
    """Cancel one or more tasks."""
    for todo in todos:
        todo.cancel()
        ctx.db.save(todo)
        click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@click.confirmation_option(prompt="Are you sure you want to delete all done tasks?")
@catch_errors
def flush(ctx: AppContext) -> None:
    """
    Delete done tasks. This will also clear the cache to reset task IDs.
    """
    database = ctx.db
    for todo in database.flush():
        click.echo(ctx.formatter.simple_action("Flushing", todo))


@cli.command()
@pass_ctx
@click.argument("ids", nargs=-1, required=True, type=click.IntRange(0))
@click.option("--yes", is_flag=True, default=False)
@catch_errors
def delete(ctx: AppContext, ids: list[int], yes: bool) -> None:
    """
    Delete tasks.

    Permanently deletes one or more task. It is recommended that you use the
    `cancel` command if you wish to remove this from the pending list, but keep
    the actual task around.
    """

    todos = []
    for i in ids:
        todo = ctx.db.todo(i)
        click.echo(ctx.formatter.compact(todo))
        todos.append(todo)

    if not yes:
        click.confirm("Do you want to delete those tasks?", abort=True)

    for todo in todos:
        click.echo(ctx.formatter.simple_action("Deleting", todo))
        ctx.db.delete(todo)


@cli.command()
@pass_ctx
@click.option(
    "--list", "-l", callback=_validate_list_param, help="The list to copy the tasks to."
)
@click.argument("ids", nargs=-1, required=True, type=click.IntRange(0))
@catch_errors
def copy(ctx: AppContext, list: TodoList, ids: list[int]) -> None:
    """Copy tasks to another list."""

    for id in ids:
        original = ctx.db.todo(id)
        todo = original.clone()
        todo.list = list
        click.echo(ctx.formatter.compact(todo))
        ctx.db.save(todo)


@cli.command()
@pass_ctx
@click.option(
    "--list", "-l", callback=_validate_list_param, help="The list to move the tasks to."
)
@click.argument("ids", nargs=-1, required=True, type=click.IntRange(0))
@catch_errors
def move(ctx: AppContext, list: TodoList, ids: list[int]) -> None:
    """Move tasks to another list."""

    for id in ids:
        todo = ctx.db.todo(id)
        click.echo(ctx.formatter.compact(todo))
        assert todo.list, "Source todo must have a list"
        ctx.db.move(todo, new_list=list, from_list=todo.list)


@cli.command(name="list")
@pass_ctx
@click.argument("lists", nargs=-1, callback=_validate_lists_param)
@click.option("--location", help="Only show tasks with location containg TEXT")
@click.option("--grep", help="Only show tasks with message containg TEXT")
@click.option(
    "--sort",
    help=(
        "Sort tasks using fields like : "
        '"start", "due", "priority", "created_at", "percent_complete" etc.'
        "\nFor all fields please refer to: "
        "<https://todoman.readthedocs.io/en/stable/usage.html> "
    ),
    callback=_sort_callback,
)
@click.option(
    "--reverse/--no-reverse",
    default=True,
    help="Sort tasks in reverse order (see --sort). Defaults to true.",
)
@click.option(
    "--due", default=None, help="Only show tasks due in INTEGER hours", type=int
)
@click.option(
    "--category",
    "-c",
    multiple=True,
    default=(),
    help="Only show tasks with specified categories.",
    callback=_validate_categories_param,
)
@click.option(
    "--priority",
    default=None,
    help=(
        "Only show tasks with priority at least as high as TEXT (low, medium or high)."
    ),
    type=str,
    callback=_validate_priority_param,
)
@click.option(
    "--start",
    default=None,
    callback=_validate_start_date_param,
    nargs=2,
    help="Only shows tasks before/after given DATE",
)
@click.option(
    "--startable",
    default=None,
    is_flag=True,
    callback=_validate_startable_param,
    help=(
        "Show only todos which "
        "should can be started today (i.e.: start time is not in the "
        "future)."
    ),
)
@click.option(
    "--status",
    "-s",
    default="NEEDS-ACTION,IN-PROCESS",
    callback=validate_status,
    help=(
        "Show only todos with the "
        "provided comma-separated statuses. Valid statuses are "
        '"NEEDS-ACTION", "CANCELLED", "COMPLETED", "IN-PROCESS" or "ANY"'
    ),
)
@catch_errors
def list_command(ctx: AppContext, *args, **kwargs) -> None:
    """
    List tasks (default). Filters any completed or cancelled tasks by default.

    If no arguments are provided, all lists will be displayed, and only
    incomplete tasks are show. Otherwise, only todos for the specified list
    will be displayed.

    eg:
      \b
      - `todo list' shows all unfinished tasks from all lists.
      - `todo list work' shows all unfinished tasks from the list `work`.

    This is the default action when running `todo'.

    The following commands can further filter shown todos, or include those
    omited by default:
    """
    hide_list = (len(list(ctx.db.lists())) == 1) or (len(kwargs["lists"]) == 1)

    kwargs["categories"] = kwargs.pop("category")

    todos = ctx.db.todos(**kwargs)
    click.echo(ctx.formatter.compact_multiple(todos, hide_list))
